@@ -41,11 +41,14 @@ module Control.Monad.Exception.Synchronous (
    assert,
    catch,
    resolve,
+   merge,
 
    ExceptionalT(..),
+   fromMaybeT,    toMaybeT,
    fromErrorT,    toErrorT,
    fromEitherT,   toEitherT,
    fromExitCodeT, toExitCodeT,
+   switchT,
    forceT,
    mapExceptionT,
    mapExceptionalT,
@@ -57,13 +60,15 @@ module Control.Monad.Exception.Synchronous (
    tryT,
    manyT,
    manyMonoidT,
+   mergeT,
    ) where
 
 import Control.Applicative (Applicative(pure, (<*>)))
-import Control.Monad (liftM, {- MonadPlus(mzero, mplus), -})
+import Control.Monad (liftM, liftM2, {- MonadPlus(mzero, mplus), -})
 import Control.Monad.Fix (MonadFix, mfix, )
 import Control.Monad.Trans.Class (MonadTrans, lift, )
 import Control.Monad.Trans.Error (ErrorT(ErrorT, runErrorT))
+import Control.Monad.Trans.Maybe (MaybeT(MaybeT, runMaybeT))
 import Data.Monoid(Monoid, mappend, mempty, Endo(Endo, appEndo), )
 
 import System.Exit (ExitCode(ExitSuccess, ExitFailure), )
@@ -185,6 +190,33 @@ resolve handler x =
       Exception e -> handler e
 
 
+
+-- like Applicative.<*>
+infixl 4 `merge`, `mergeT`
+
+{- | see 'mergeT' -}
+merge, mergeLazy, _mergeStrict ::
+   (Monoid e) =>
+   Exceptional e (a -> b) -> Exceptional e a -> Exceptional e b
+merge = mergeLazy
+
+mergeLazy ef ea =
+   case ef of
+      Exception e0 ->
+         Exception $ mappend e0 $
+         case ea of
+            Success _ -> mempty
+            Exception e1 -> e1
+      Success f -> fmap f ea
+
+_mergeStrict ef ea =
+   case (ef,ea) of
+      (Success f, Success a) -> Success $ f a
+      (Exception e, Success _) -> Exception e
+      (Success _, Exception e) -> Exception e
+      (Exception e0, Exception e1) -> Exception $ mappend e0 e1
+
+
 instance Functor (Exceptional e) where
    fmap f x =
       case x of
@@ -232,6 +264,15 @@ newtype ExceptionalT e m a =
    ExceptionalT {runExceptionalT :: m (Exceptional e a)}
 
 
+_assertMaybeT :: (Monad m) => e -> Maybe a -> ExceptionalT e m a
+_assertMaybeT e = maybe (throwT e) return
+
+fromMaybeT :: Monad m => e -> MaybeT m a -> ExceptionalT e m a
+fromMaybeT e  =  ExceptionalT . liftM (fromMaybe e) . runMaybeT
+
+toMaybeT :: Monad m => ExceptionalT e m a -> MaybeT m a
+toMaybeT  =  MaybeT . liftM toMaybe . runExceptionalT
+
 fromErrorT :: Monad m => ErrorT e m a -> ExceptionalT e m a
 fromErrorT  =  fromEitherT . runErrorT
 
@@ -256,6 +297,14 @@ fromExitCodeT ::
 fromExitCodeT act =
    ExceptionalT $ fmap fromExitCode act
 
+
+switchT ::
+   (Monad m) =>
+   (e -> m b) -> (a -> m b) ->
+   ExceptionalT e m a -> m b
+switchT e s m =
+   switch e s =<< runExceptionalT m
+
 {- |
 see 'force'
 -}
@@ -357,6 +406,26 @@ manyMonoidT handler action =
    in  recourse
 
 
+{- |
+This combines two actions similar to Applicative's @<*>@.
+The result action fails if one of the input action fails,
+but both actions are executed.
+E.g. consider a compiler that emits all errors
+that can be detected independently,
+but eventually aborts if there is at least one error.
+
+The exception type @e@ might be a list type,
+or an @Endo@ type that implements a difflist.
+-}
+mergeT ::
+   (Monoid e, Monad m) =>
+   ExceptionalT e m (a -> b) ->
+   ExceptionalT e m a ->
+   ExceptionalT e m b
+mergeT mf ma =
+   ExceptionalT $
+   liftM2 merge (runExceptionalT mf) (runExceptionalT ma)
+
 
 instance Functor m => Functor (ExceptionalT e m) where
    fmap f (ExceptionalT x) =
@@ -377,7 +446,7 @@ instance Monad m => Monad (ExceptionalT e m) where
             Success x -> runExceptionalT $ f x
 
 instance (MonadFix m) => MonadFix (ExceptionalT e m) where
-   mfix f = ExceptionalT $ mfix $ \(Success r) -> runExceptionalT $ f r
+   mfix f = ExceptionalT $ mfix $ \ ~(Success r) -> runExceptionalT $ f r
 
 instance MonadTrans (ExceptionalT e) where
    lift m = ExceptionalT $ liftM Success m
